# -*- coding: utf-8 -*-
"""
This file contains several functions used in the ACERIM workflow. They are 
sorted into the following sections:

- ACERIM FUNCTIONS - top-level functions
- PLOTTING - plot figures
- ROI MANIPULATION - manipulate 2D arrays of image data
- GEOSPATIAL CALCULATIONS - manipulate geospatial data
- STATISTICS - compute statistics and fitting functions

For usage, see sample/tutorial.rst.
"""
from __future__ import division, print_function, absolute_import
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from acerim import acestats as acs

######################### ACERIM FUNCTIONS ####################################
def ejecta_profile_stats(cdf, ads, ejrad=2, rspacing=0.25, stats=None, 
                         plot_roi=False, vmin=None, vmax=None, strict=False, 
                         plot_stats=False, savepath=None, timer=False, dt=60):
    """Compute stat profiles across ejecta blankets by computing stats in a 
    series of concentric rings beginning at the crater rim and extending to 
    ejrad crater radii from the crater centre. Each ring is rspacing thick and
    the crater floor is excluded. Output is a dictionary of cdf.index values to
    pandas.DataFrame of stats containing the computed stats as columns and 
    ring radius as rows.
    
    Parameters
    ----------
    
    Returns
    -------
    stat_df : dict of (index: pandas.DataFrame)
        Dictionary of stats using the cdf index as keys and the pandas.DataFrame
        of statistics as values (see example)
    
    Example
    -------
    >>> compute_ejecta_profile_stats(my_cdf, my_ads, 2, 0.25, stats=['mean',
                                    'median','maximum'],savepath='/tmp/')
    
    /tmp/index1_ejecta_stats.csv
        radius,mean,median,maximum
        1,     55,  45,    70
        1.25,  50,  40,    65
        1.5,   35,  35,    45
        1.75,  30,  33,    42
        
    """ 
    if timer:
        from timeit import default_timer as timer
        Tstart, Tnow = timer(), timer()
        count = 1 # count craters
    if savepath:
        import os.path as p
        savename = p.join(savepath,'{}_ejpstats.csv')
    if stats is None:
        stats = acs._listStats()
    stat_dict = {}
    for ind in cdf.index: # Main CraterDataFrame loop
        lat = cdf.loc[ind, cdf.latcol]
        lon = cdf.loc[ind, cdf.loncol]
        rad = cdf.loc[ind, cdf.radcol]
        try:
            roi, extent = ads.get_roi(lat, lon, rad, ejrad, get_extent=True)
            ring_array = np.arange(1, ejrad+rspacing, rspacing) # inner ring radius [crater radii]
            ring_darray = ring_array*rad # inner ring radius [km]
            stat_df = pd.DataFrame(index=ring_array[:-1], columns=stats)
            for i in range(len(ring_array)-1): # Ejecta ring radii loop  
                rad_inner = ring_darray[i] # inner radius of this ejecta ring 
                rad_outer = ring_darray[i+1] # outer radius of this ejecta ring
                mask = crater_ring_mask(ads, roi, lat, lon, rad_inner, rad_outer)
                roi_masked = mask_where(roi, ~mask)
                filtered_roi = filter_roi(roi_masked, vmin, vmax, strict)
                roi_notnan = filtered_roi[~np.isnan(filtered_roi)]
                for stat, function in acs._getFunctions(stats):
                    stat_df.loc[ring_array[i], stat] = function(roi_notnan)                
                if plot_roi:
                    ads.plot_roi(filtered_roi, extent=extent) 
            stat_dict[ind] = stat_df
            if plot_stats:
                plot_ejecta_stats(stat_df)
            if savepath:
                stat_df.to_csv(savename.format(ind))
        except ImportError as e: # Catches and prints out of bounds exceptions
            print(e, ". Skipping....")
        if timer: # Print a status update approx every dt seconds
            if (timer() - Tnow > dt) or (count == len(cdf)):
                update = 'Finished crater {} out of {}'.format(count, len(cdf))
                elapsed = timer()-Tstart
                update += '\n  Time elapsed: {}h:{}m:{}s'.format(int(elapsed//3600),
                              int((elapsed%3600)//60), round((elapsed%3600)%60,2))
                print(update)
                Tnow = timer()
            count +=1
    return stat_dict
    

def ejecta_stats(cdf, ads, ejrad=2, stats=None, plot=False, vmin=None, 
                 vmax=None, strict=False):
    """Compute the specified stats from acestats.py on a circular ejecta ROI
    extending ejsize crater radii from the crater center. Return results as a 
    CraterDataFrame the same length as cdf with stats appended as extra columns.    
    
    Parameters
    ---------- 
    cdf : CraterDataFrame object
        Contains the crater locations and sizes to locate ROIs in aceDS. Also
        form the basis of the returned CraterDataFrame
    ads : AceDataset object
        AceDataset of image data used to compute stats.
    ejrad : int, float
        The radius of ejecta blanket measured in crater radii from the 
        crater center to the ejecta extent.
    stats : Array-like of str
        Indicates the stat names to compute. Stat functions must be defined in 
        acestats.py. Default: all stats in acestats.py. 
    plot : bool
        Plot the ejecta ROI.
    vmin : int, float
        The minimum valid image data value. Set this to filter all values lower.
    vmax : int, float
        The maximum valid image data value. Set this to filter all values higher.
    strict : bool
        How strict the (vmin, vmax) range is. If true, exclude values <= vmin
        and values >= vmax, if they are specified.

    Returns
    -------
    CraterDataFrame
        Same length as cdf with stats included as new columns.
    """
    # If stats and index not provided, assume use all stats and all rows in cdf
    if stats is None:
        stats = acs._listStats()
    # Initialize return CraterDataframe with stats as individual columns
    ret_cdf = cdf
    for stat in stats:
        ret_cdf[stat] = ret_cdf.index
    # Main computation loop
    for i in cdf.index:
        # Get lat, lon, rad and compute roi for current crater
        lat = cdf.loc[i, cdf.latcol]
        lon = cdf.loc[i, cdf.loncol]
        rad = cdf.loc[i, cdf.radcol]
        roi, extent = ads.get_roi(lat, lon, rad, ejrad, get_extent=True)
        mask = crater_ring_mask(ads, roi, lat, lon, rad, rad*ejrad)
        roi_masked = mask_where(roi, ~mask) # ensure it is ~mask to keep the ring interior
        filtered_roi = filter_roi(roi_masked, vmin, vmax, strict)
        roi_notnan = filtered_roi[~np.isnan(filtered_roi)] # Collapse to 1D non-nan array
        for stat, function in acs._getFunctions(stats):
            ret_cdf.loc[i, stat] = function(roi_notnan)
        if plot: # plot filtered and masked roi
            ads.plot_roi(filtered_roi, extent=extent)
    return ret_cdf
  
   
################################ PLOTTING #####################################
def plot_roi(ads, roi, figsize=((8,8)), extent=None, title='ROI', vmin=None,
             vmax=None, cmap='gray', **kwargs):
    """
    Plot roi 2D array. 
    
    If extent, cname and cdiam are supplied, the axes will display the 
    lats and lons specified and title will inclue cname and cdiam.
    
    Parameters
    ----------
    ads : AceDataset object
        The parent dataset of roi.
    roi : 2D array
        The roi from ads to plot.
    figsize : tuple
        The (length,width) of plot in inches.
    extent : array-like
        The [minlon, maxlon, minlat, maxlat] extents of the roi in degrees.
    title : str
        Title of the roi.
    vmin : int, float
        Minimum pixel data value for plotting.
    vmax : int, float
        Maximum pixel data value for plotting.
    cmap : str
        Color map name for plotting. Must be a valid colorbar in
        matplotlib.cm.cmap_d. See help(matplotlib.cm) for full list.
        
    Other parameters
    ----------------
    **kwargs : object
        Additional keyword arguments to be passed to the imshow function. See 
        help(matplotlib.pyplot.imshow) for more info.
    """
    plt.figure("ROI", figsize=figsize)
    plt.imshow(roi, extent=extent, cmap=cmap, vmin=vmin, vmax=vmax)
    plt.title(title)
    plt.xlabel('Longitude (degrees)')
    plt.ylabel('Latitude (degrees)')
    plt.show()


########################### ROI MANIPULATION ##################################
def filter_roi(roi, vmin=None, vmax=None, strict=False):
    """
    Filter values outside (vmin, vmax) by setting them to np.nan. If strict,
    keep only values strictly less than vmax and strictly greater than vmin.
    
    E.g. strict=False keeps vmin <= roi <= vmax
         strict=True keeps vmin < roi < vmax
    """
    if vmin is None:
        vmin = -np.inf
    if vmax is None:
        vmax = np.inf
    nanmask = ~np.isnan(roi) # build nanmask with pre-existing nans, if any
    nanmask[nanmask] &= roi[nanmask] > vmax # Add values outside of (vmin,vmax) to nanmask
    nanmask[nanmask] &= roi[nanmask] < vmin
    if strict: # if strict, also exclude values equal to vmin, vmax
        nanmask[nanmask] &= roi[nanmask] == vmax
        nanmask[nanmask] &= roi[nanmask] == vmin
    roi[nanmask] = np.nan
    return roi

def mask_where(ndarray, condition):
    """
    Return copy of ndarray with nan entries where condition is True.
    
    >>> arr = np.array([1.,2.,3.,4.,5.])
    >>> masked = mask_where(arr, arr > 3) 
    >>> np.isnan(masked[3:]).all()
    True
    """
    mask = np.array(np.ones(ndarray.shape)) # Same shape array of ones
    mask[np.where(condition)] = np.nan 
    return ndarray * mask


def circle_mask(roi, radius, center=(None,None)):
    """
    Return boolean array of True inside circle of radius at center.
    
    >>> roi = np.ones((3,3))
    >>> masked = circle_mask(roi, 1)
    >>> masked[1,1]
    True
    >>> masked[0,0]
    False
    """
    if not center[0]: # Center circle on center of roi
        center = np.array(roi.shape)/2 - 0.5
    cx, cy = center
    width, height = roi.shape
    x = np.arange(width) - cx
    y = np.arange(height).reshape(-1,1) - cy
    if radius > 0:
        return x*x + y*y <= radius*radius
    else: 
        return np.zeros(roi.shape, dtype=bool)


def ellipse_mask(roi, a, b, center=(None, None)):
    """
    Return boolean array of True inside ellipse with horizontal major axis a 
    and vertical minor axis b centered at center.
    
    >>> roi = np.ones((9,9))
    >>> masked = ellipse_mask(roi, 3, 2)
    >>> masked[4,1]
    True
    >>> masked[4,0]
    False
    """
    if not center[0]: # Center ellipse on center of roi
        center = np.array(roi.shape)/2 - 0.5
    cx, cy = center    
    width, height = roi.shape
    y, x = np.ogrid[-cx:width-cx, -cy:height-cy]
    return (x*x)/(a*a) + (y*y)/(b*b) <= 1


def ring_mask(roi, rmin, rmax, center=(None,None)):
    """
    Return boolean array of True in a ring from rmin to rmax radius around 
    center. Returned array is same shape as roi.
    """
    inner = circle_mask(roi, rmin, center)
    outer = circle_mask(roi, rmax, center)
    return outer*~inner


def crater_floor_mask(aceds, roi, lat, lon, rad):
    """
    Mask the floor of the crater lat, lon, with radius rad.
    """
    pixwidth = m2pix(rad, aceds.calc_mpp(lat))
    pixheight = m2pix(rad, aceds.calc_mpp())
    return ellipse_mask(roi, pixwidth, pixheight)


def crater_ring_mask(aceds, roi, lat, lon, rmin, rmax):
    """
    Mask a ring around a crater with inner radius rmin and outer radius rmax
    crater radii.
    """ 
    rmax_pixheight = m2pix(rmax, aceds.calc_mpp())
    rmax_pixwidth = m2pix(rmax, aceds.calc_mpp(lat))
    rmin_pixheight = m2pix(rmin, aceds.calc_mpp())
    rmin_pixwidth = m2pix(rmin, aceds.calc_mpp(lat))
    outer = ellipse_mask(roi, rmax_pixwidth, rmax_pixheight) 
    inner = ellipse_mask(roi, rmin_pixwidth, rmin_pixheight)
    return outer * ~inner


######################### GEOSPATIAL CALCULATIONS #############################
def inbounds(lat, lon, mode='std'):
    """True if lat and lon within global coordinates.
    Standard: mode='std' for lat in (-90, 90) and lon in (-180, 180).
    Positive: mode='pos' for lat in (0, 180) and lon in (0, 360)
    
    >>> lat = -10
    >>> lon = -10
    >>> inbounds(lat, lon)
    True
    >>> inbounds(lat, lon, 'pos')
    False
    """
    if mode == 'std':
        return (-90 <= lat <= 90) and (-180 <= lon <= 180)
    elif mode == 'pos':
        return (0 <= lat <= 180) and (0 <= lon <= 360)
    
def m2deg(dist, mpp, ppd):
    """Return dist converted from meters to degrees."""
    return dist/(mpp*ppd)

def m2pix(dist, mpp):
    """Return dist converted from meters to pixels"""
    return int(dist/mpp)

def deg2pix(dist, ppd):
    """Return dist converted from degrees to pixels."""
    return int(dist*ppd)  


def get_ind(value, array):
    """Return closest index (rounded down) of a value from sorted array."""
    ind = np.abs(array-value).argmin() 
    return int(ind)   


def deg2rad(theta):
    """
    Convert degrees to radians.
    
    >>> deg2rad(180)
    3.141592653589793
    """
    return theta * (np.pi / 180)


def greatcircdist(lat1, lon1, lat2, lon2, radius):
    """
    Return great circle distance between two points on a spherical body.
    Uses Haversine formula for great circle distances.
    
    >>> greatcircdist(36.12, -86.67, 33.94, -118.40, 6372.8)
    2887.259950607111
    """
    if not all(map(inbounds,(lat1,lon1),(lat2,lon2))) or abs(lat1)==90 or abs(lat2)==90:
        raise ValueError("Latitude or longitude out of bounds.")
    # Convert degrees to radians
    lat1, lon1, lat2, lon2 = map(deg2rad, [lat1, lon1, lat2, lon2])
    # Haversine
    dlat = lat2 - lat1
    dlon = lon2 - lon1
    a = np.sin(dlat/2)**2 + np.cos(lat1)*np.cos(lat2)*np.sin(dlon/2)**2
    theta = 2 * np.arcsin(np.sqrt(a))
    dist = radius*theta
    return dist


############################### STATISTICS ####################################
#import numpy as np
#import scipy.optimize as opt
#import helper_functions as hf

def histogram(roi, bins, hmin=None, hmax=None, skew=False, *args, **kwargs):
    """
    Return histogram, bins of histogram computed on ROI. See np.histogram for 
    full usage and optional parameters. Set verbose=True to print a summary of
    the statistics.
    """
    roi_notnan = roi[~np.isnan(filtered_roi)]
    roi_valid = roi_notnan[hmin <= roi_notnan <= hmax]
    hist, bins = np.histogram(roi, bins=bins, hmin=hmin, hmax=hmax, *args, **kwargs)
    ret = [hist, bins]
    output = 'Histogram with {} pixels total, \
              {} pixels in [hmin, hmax] inclusive, \
              {} nan pixels excluded, \
              {} bins'.format(len(roi), len(roi_valid), len(roi_notnan), len(bins))
    if skew:
        skewness = scipy.stats.skew(hist)
        ret.append(skewness)
        output += ', {} skewness'.format(skewness)
    if verbose:
        print(output)
    return ret


# Ejecta profile stats

def fit_exp(x, y, PLOT_EXP=False):
    """
    Return an exponential that has been fit to data using scipy.curvefit(). 
    If plot is True, plot the data and the fit. 
    """     
    def exp_eval(x, a, b, c):
        """
        Return exponential of x with a,b,c parameters.
        """
        return a * np.exp(-b * x) + c      
      
    try:
        p_opt, cov = opt.curve_fit(expEval, x, y)
    except:
        RuntimeError
        return None
    if PLOT_EXP:
        hf.plot_exp(x, y, exp_eval(x, *p_opt))
    return p_opt


def fit_gauss(data, PLOT_GAUSS=False):
    """
    Return parameters of a Gaussian that has been fit to data using least 
    squares fitting. If plot=True, plot the histogram and fit of the data. 
    """
    def gauss(x, p): 
        """
        Return Gaussian with mean = p[0], standard deviation = p[1].
        """
        return 1.0/(p[1]*np.sqrt(2*np.pi))*np.exp(-(x-p[0])**2/(2*p[1]**2))
   
    data = data[data > 0]       
    n, bins = np.histogram(data, bins='fd', density=True) 
    p0 = [0, 1] #initial parameter guess
    errfn = lambda p, x, y: gauss(x, p) - y
    p, success = opt.leastsq(errfn, p0[:],args=(bins[:-1], n))
    if PLOT_GAUSS:
        hf.plot_gauss(bins, n, gauss(bins, p))
    return p
    
    
def fit_pow(xdata, ydata, p0, plot=False, cid=''):
    """
    Return a power law curve fit of y using a least squares fit.
    """
    def residuals(p, x, y):
        return ydata - pow_eval(p, x)
    
    def pow_eval(p, x):
        return p[0] + p[1] * (x**p[2])

    pfinal, success = opt.leastsq(residuals, p0, args=(xdata, ydata))
    xarr = np.arange(xdata[0], xdata[-1], 0.1)
    return xarr, pow_eval(pfinal, xarr)
    
    
def fit_pow_linear(xdata,ydata,p0,plot=False,cid=''):
    """
    Return a power law curve fit of y by converting to linear data using 
    logarithms and then performing a linear least squares fit.
    """
    def fit_line(p, x):
        return p[0] + p[1] * x        
        
    def residuals(p, x, y):
        return ydata - fitLine(p,x)
    
    def pow_eval(x,amp,index):
        return amp * (x**index)
        
    logx = np.log(xdata)
    logy = np.log(ydata)
    pfinal, success = opt.leastsq(residuals, p0, args=(logx, logy))
    
    return pow_eval(xdata, pfinal[0], pfinal[1])
    
#%%
#if __name__ == "__main__":
#    import doctest
#    doctest.testmod()
    



"""
Welcome to the ACERIM tutorial!
===============================

**IN PROGRESS**

This will walk you through an example of a research script using ACERIM to 
introduce and explain the various features of the package. Full API 
documentation is also available at www.readthedos.org/projects/acerim.
"""

""" Housekeeping ""
For this tutorial to run on your system without you specifying a path to your 
ACERIM installation, we will need to import the built-in module "os". You 
should not require os for your general ACERIM workflow because you can simply
specify the path to your directory.
"""
import os.path


""" Importing classes from ACERIM """
"""A typical acerim workflow will begin by importing the 2 main ACERIM classes,
the AceDataset and the CraterDataFrame. Here we shorten these to Ads and Cdf, 
respectively, to avoid typing acerim.aceclasses.AceDataset each time we want to
load an image.
"""
from acerim import AceDataset as Ads
from acerim import CraterDataFrame as Cdf


""" Path to your data """
"""In general, you can specify data_dir directly with your desired file path.
"""
data_dir = '/path/to/your/data'

"""Here, we will use os to find this file's directory (where our sample data is
located):
"""
sample_dir = os.path.dirname(os.path.abspath(__file__))


""" The AceDataset """

""" The AceDataset class is used to import and manipulate image data. Note that 
acerim assumes that supplied image data is in a simple cylindrical projection. 
If a geotiff is supplied, the AceDataset will attempt to read and store the 
geographical information automatically. Importing moon.tif from image_data_path:
"""
moon_image_path = sample_dir + '/moon.tif'
moon = Ads(moon_image_path)
print(moon)

""" Printing the AceDataset (called "moon") is a good way to test that the image 
was imported, but it also gives a breakdown of the geospatial information 
it contains. In this case, moon printed the wrong radius for the Moon (it 
should be 1737.4 km, not 6378 km). All geospatial information is saved
in moon's attributes. In this case the offending attribute can be checked with: 
"""
moon.radius

""" Since moon has the wrong radius, we will have to make a new AceDataset, 
this time specifying the correct radius. Any of the geospatial parameters can 
be set by specifying them in the constructor:
"""
moon = Ads(moon_image_path, nlat=90, slat=-90, wlon=-180, elon=180, 
		   radius=1737.4, ppd=4)
print(moon)

""" The main feature of the AceDataset is extracting 2D image arrays (dubbed 
ROIs) from crater locations in the dataset. Say we want to see Copernicus 
crater on the Moon. We will need to supply the AceDataset.get_roi method with
the latitude, longitude and radius of Copernicus.
"""
cop_lat, cop_lon, cop_rad = [9.62, 20.08, 93]

""" We will also need to supply a window size (in crater radii). Let's say we 
want to see 5 crater radii from the center of the crater:
"""
cop_wsize = 5

""" Now we have all we need to extract an ROI from moon:
"""
cop_roi = moon.get_roi(cop_lat, cop_lon, cop_rad, cop_wsize)

""" ROIs take the form of numpy 2D arrays. We can check the size of our ROI
with:
"""
cop_roi.shape

""" Now let's take a look at Aristarchus, but plot the ROI on the fly:
"""
arist_roi = moon.get_roi(23.7, -47.4, 40, 5, plot_roi=True)

""" To have more plotting options you can pass the roi to the plot_roi method 
of AceDataset:
"""
moon.plot_roi(arist_roi, figsize=(5,5), title='Aristarchus', vmin=0, vmax=75,
			  cmap='jet')


""" The CraterDataFrame """

""" In this section we will explore how to use the CraterDataFrame to store and
access crater data in this tabular data structure.
"""
crater_sheet_path = sample_dir + '/craters.csv'

""" Importing a crater spreadsheet """
""" A CraterDataFrame can be initialized by passing a string to the 
constructor. Doing so assumes that the string is a file path to a csv (comma 
separated values) file:
"""
craters_fromcsv = Cdf(crater_sheet_path)

""" Data can also be read from a properly formatted dictionary:
"""
cdict = {'Lat' : [10, -20., 80.0], 
         'Lon' : [14, -40.1, 317.2],
         'Diam' : [2, 12., 23.7]}
craters_fromdict = Cdf(cdict)

""" The CraterDataFrame also accepts pandas.DataFrame objects, so that if 
other import options are needed (importing from an Excel spradsheet, specifying
index columns, etc), this can be handled with pandas and then passed to the
CraterDataFrame like so:
"""
import pandas
dataframe = pandas.read_csv(crater_sheet_path)
craters_frompandas = Cdf(dataframe)

""" A final import option that is useful for simplying queries is the index
option. If you have a column in your data with crater names or unique indices,
you can specify its position to make it the default DataFrame index. In this 
case it is the first column of data (position 0 in python):
"""
craters = Cdf(crater_sheet_path, index_col=0)

""" Visualizing and Querying the CraterDataFrame """
""" Methods that are available to the pandas.DataFrame are also available to
the CraterDataFrame. For instance to give a quick check of the first few lines 
of data, we can use the pandas.DataFrame method "head":
"""
craters.head()

""" To query a crater assuming we set the index to the Name column, we use 
loc['Crater Name']:
"""
craters.loc['Humboldt']

""" There are different way to get a value from the CraterDataFrame. This will 
return the longitude of the crater:
"""
craters.loc['Humboldt']['Lon']

""" But using ".at[]" is the most efficient way to get a single cell of data:
"""
craters.at['Humboldt', 'Lon']

""" DataFrames also support fancy (boolean) indexing. To get all crater larger 
than 9 km in diameter:
"""
craters[craters['Diam'] > 9]

""" Or to get all craters with longitude between 0 and 10 degrees:
"""
craters[(craters['Lon'] > 0) & (craters['Lon'] < 10)]

""" Conditions ca be strung together to filter your data:
"""
craters = craters[(craters['Lat'] > -80) & (craters['Lat'] < 80)]


""" Advanced Examples """

""" Loop through the first 5 names in the craters index, 
extract the lat, lon, radius and plot the ROIs
"""
for name in craters.index[:5]:
    lat = craters.at[name, 'Lat']
    lon = craters.at[name, 'Lon']
    rad = craters.at[name, '_Rad']
    roi = moon.get_roi(lat, lon, rad, wsize=4, plot_roi=True)

        
""" The acestats module """

""" Statistics can be performed on an roi using the acestats module.
"""
# TODO: acestats examples
    
    



# -*- coding: utf-8 -*-
"""
Suite of unittests for functions found in /acerim/acefunctions.py.
"""
from __future__ import division, print_function, absolute_import
import os
import unittest
import numpy as np
import acerim
from acerim import acefunctions as af
from acerim import aceclasses as ac

DATA_PATH = os.path.join(acerim.__path__[0], 'sample')

#%% Test Compute Stats
class Test_compute_stats(unittest.TestCase):
    """Test computeStats function"""
    crater_csv = os.path.join(DATA_PATH,'craters.csv')
    cdf = ac.CraterDataFrame(crater_csv)
    test_dataset = os.path.join(DATA_PATH, 'moon.tif')
    ads = ac.AceDataset(test_dataset, radius=1737)

    def test_first_mean(self):
        """Test mean on first crater in cdf"""
        #af.compute_stats(self.cdf, self.ads, 'mean', self.cdf.index[0:5])
        # TODO: fix this test (choose appropriate indices)

#%% Test ROI manipulation functions
class Test_circle_mask(unittest.TestCase):
    """Test ring_mask function"""
    def test_trivial(self):
        """Test radius 0"""
        actual = af.circle_mask(np.ones((3,3)), 0)
        expected = np.array([[False, False, False],
                             [False, False, False],
                             [False,  False, False]])
        self.assertIsNone(np.testing.assert_array_equal(actual, expected))
    
    def test_odd(self):
        """Test roi with odd side length"""
        actual = af.circle_mask(np.ones((5,5)), 2)
        expected = np.array([[False, False, True,  False, False],
                             [False, True,  True,  True,  False],
                             [True,  True,  True,  True, True],
                             [False, True,  True,  True,  False],
                             [False, False, True,  False, False]])
        self.assertIsNone(np.testing.assert_array_equal(actual, expected))
        
    def test_even(self):
        """Test roi with even side length"""
        actual = af.circle_mask(np.ones((4,4)), 2)
        expected = np.array([[False,  True,  True, False],
                             [ True,  True,  True,  True],
                             [ True,  True,  True,  True],
                             [False,  True,  True, False]])
        self.assertIsNone(np.testing.assert_array_equal(actual, expected))
        
    def test_offcenter(self):
        """Test specifying off center location"""
        actual = af.circle_mask(np.ones((5,5)), 2, center=(3,2))
        expected = np.array([[False, False, False, True,  False],
                             [False, False, True,  True,  True],
                             [False, True,  True,  True,  True],
                             [False, False, True,  True,  True],
                             [False, False, False, True,  False]])
        self.assertIsNone(np.testing.assert_array_equal(actual, expected))
     

class Test_ring_mask(unittest.TestCase):
    """Test ring_mask function"""
    def test_trivial(self):
        actual = af.ring_mask(np.ones((3,3)), 0, 0)
        expected = np.array([[False, False, False],
                             [False, False, False],
                             [False,  False, False]])
        self.assertIsNone(np.testing.assert_array_equal(actual, expected))
    
    def test_odd(self):
        """Test roi with odd side length"""
        actual = af.ring_mask(np.ones((5,5)), 1, 2)
        expected = np.array([[False, False, True,  False, False],
                             [False, True,  False, True,  False],
                             [True,  False, False, False, True],
                             [False, True,  False, True,  False],
                             [False, False, True,  False, False]])
        self.assertIsNone(np.testing.assert_array_equal(actual, expected))
        
    def test_even(self):
        """Test roi with even side length"""
        actual = af.ring_mask(np.ones((4,4)), 1.5, 2)
        expected = np.array([[False, True,  True,  False],
                             [True,  False, False, True],
                             [True,  False, False, True],
                             [False, True,  True,  False]])
        self.assertIsNone(np.testing.assert_array_equal(actual, expected))
        
    def test_offcenter(self):
        """Test specifying off center location"""
        actual = af.ring_mask(np.ones((5,5)), 1, 2, center=(3,2))
        expected = np.array([[False, False, False, True,  False],
                             [False, False, True,  False, True],
                             [False, True,  False, False, False],
                             [False, False, True,  False, True],
                             [False, False, False, True,  False]])
        self.assertIsNone(np.testing.assert_array_equal(actual, expected))
         

#%% Image Helper Functions
class Test_m2deg(unittest.TestCase):
    """Test m2deg functions"""
    def test_basic(self):
        """Test simple"""
        actual = af.m2deg(400, 10, 20)
        expected = 2.0
        self.assertEqual(actual, expected)
    
    def test_float(self):
        """Test float"""
        actual = af.m2deg(1500.0, 4.0, 0.25)
        expected = 1500.0
        self.assertEqual(actual, expected)
        


""" 
Keeps track of the current acerim version. Versions are specified as a string
of the form "X.Y.Z" (major.minor.maintenance). Versions that are still in 
development append 'dev0' to the end (e.g., "X.Y.Zdev0").
"""
_version_major = 0
_version_minor = 1
_version_maintenance = ''  # use '' for each new major/minor release; int for 1+
_version_extra = '' # NEVER release with dev, use '' for full release


def concatenate_version(major, minor, maintenance, extra):
	""" Construct full version string to pass to setup.py """
	_ver = [major, minor]
	if maintenance:
	    _ver.append(maintenance)
	if extra:
	    _ver.append(extra)
	return '.'.join(map(str, _ver))

__version__ = concatenate_version(_version_major, _version_minor, 
									_version_maintenance, _version_extra)



# -*- coding: utf-8 -*-
"""
Suite of unittests for classes found in /acerim/aceclasses.py.
"""
from __future__ import division, print_function, absolute_import
import os
import unittest
import numpy as np
import pandas as pd
import acerim
from acerim import aceclasses as ac

DATA_PATH = os.path.join(acerim.__path__[0], 'sample')

class TestAceDataset(unittest.TestCase):
    """Test AceDataset object"""
    test_dataset = os.path.join(DATA_PATH, 'moon.tif')
    ads = ac.AceDataset(test_dataset, radius=1737)
    
    def test_file_import(self):
        """Import test tif '/tests/moon.tif'"""
        self.assertIsNotNone(self.ads)
        
    def test_get_gdalDataset_attrs(self):
        pass

    def test_get_AceDataset_attrs(self):
        pass

    def test_repr(self):
        pass
        
    def test_is_global(self):
        """Test .is_global method for checking if dataset has 360 degrees of lon""" 
        is_global = ac.AceDataset(self.test_dataset, wlon=0, elon=360).is_global()
        self.assertTrue(is_global)
        
        not_global = ac.AceDataset(self.test_dataset, wlon=0, elon=180).is_global()
        self.assertFalse(not_global)
        
    def test_calc_mpp0(self):
        """Test .calc_mpp method at equator"""
        ads = self.ads
        circum = 2*np.pi*ads.radius # [m] circumference at lat=0
        xpix = ads.RasterXSize # [pix]
        expected = circum/xpix # [m/pix]
        actual = ads.calc_mpp()
        self.assertAlmostEqual(actual, expected, 5)
        
    def test_calc_mpp50(self):
        """Test .calc_mpp method at 50 degrees latitude"""
        ads = self.ads
        circum = 2*np.pi*np.cos(50*(np.pi/180))*ads.radius
        xpix = ads.RasterXSize
        expected = circum/xpix
        actual = ads.calc_mpp(50)
        self.assertAlmostEqual(actual, expected, 5)
        
    def test_calc_mpp_gt90(self):
        """Test .calc_mpp method fails above 90 or below -90 degrees latitude"""
        ads = self.ads
        self.assertRaises(ValueError, ads.calc_mpp, 90)
        self.assertRaises(ValueError, ads.calc_mpp, 100)
        self.assertRaises(ValueError, ads.calc_mpp, -100)

    def test_get_info(self):
        """Test .get_info method for reading geotiff info"""
        ads = self.ads
        actual = ads.get_info()
        expected = (90.0, -90.0, -180.0, 180.0, 6378.137, 4.0)
        self.assertEqual(actual, expected)

    def test_get_roi(self):
        """Test get_roi method"""
        pass

    def test_wrap_lon(self):
        pass


class TestCraterDataFrame(unittest.TestCase):
    """Test CraterDataFrame object"""
    crater_csv = os.path.join(DATA_PATH,'craters.csv')
    cdict = {'Lat' : [10, -20., 80.0],
             'Lon' : [14, -40.1, 317.2],
             'Diam' : [2, 12., 23.7]}
    
    def test_file_import(self):
        """Import from test file '/tests/craters.csv'"""
        cdf = ac.CraterDataFrame(self.crater_csv)
        self.assertIsNotNone(cdf)
        
    def test_dict_import(self):
        """Import from dict"""
        cdf = ac.CraterDataFrame(self.cdict)
        self.assertIsNotNone(cdf)
    
    def test_pandas_dataframe_import(self):
        """Import from pandas.DataFrame object"""
        pdf = pd.DataFrame(pd.read_csv(self.crater_csv))
        cdf = ac.CraterDataFrame(pdf)
        self.assertIsNotNone(cdf)
        
    def test_specify_index(self):
        """Test defining custom index"""
        cdf = ac.CraterDataFrame(self.cdict, index=['A', 'B', 'C'])
        self.assertIn('A', cdf.index)
        
    def test_specify_columns(self):
        """Test defining custom index"""
        cdf = ac.CraterDataFrame(self.cdict, columns=['Diameter', 'Lat','Lon'])
        self.assertIn('Diameter', cdf.columns)   
        
    def test_find_latcol(self):
        """Find latitude column in loaded data"""
        cdf = ac.CraterDataFrame(self.cdict)
        self.assertEqual(cdf.latcol, 'Lat')
        
    def test_find_loncol(self):
        """Find longitude column in loaded data"""
        cdf = ac.CraterDataFrame(self.cdict)
        self.assertEqual(cdf.loncol, 'Lon')

    def test_make_radcol(self):
        """Make radius column if diameter is specified in loaded data"""
        cdf = ac.CraterDataFrame(self.cdict)
        self.assertEqual(cdf.radcol, 'radius')
        actual = cdf.loc[0, 'radius']
        expected = 0.5*cdf.loc[0, 'Diam']
        self.assertEqual(actual, expected)
        
        

        
        
        
        


# -*- coding: utf-8 -*-
"""
This file contains statistical functions that can be used to quickly
generate statistics on ROIs (e.g. by using the ejecta_stats or 
ejecta_profile_stats functions in acefunctions.py). 

Additional desired statistical functions can be added to this file by 
following the naming convention used here:

    def statname(roi_array):
        '''What this stat function does'''
        return statistics(roi_array)

Each function should take a single numpy array as an arugument and return a 
single value.

The private functions are:

    _listStats(): return names of all non-protected functions in this file 

    _getFunctions(stats): return array of pairs of function names and functions as 
                  specified by stats 

Non-statistical functions in this file must be private (begin with "_") to avoid
being called with other stat functions.
"""
from __future__ import division, print_function, absolute_import
import inspect
import numpy as np
import scipy as sp

# Statistical Functions
def maximum(roi):
    """Return maximum pixel value in roi"""
    return np.max(roi)

def mean(roi):
    """Return the mean of roi"""
    return np.mean(roi)

def median(roi):
    """Return the median (50/50) of roi"""
    return np.median(roi)

def minimum(roi):
    """Return minimum pixel value in roi"""
    return np.min(roi)

def pct95(roi):
    """Return the 95th percentile (95/5) value of roi"""
    return np.percentile(roi, 95)

def q1(roi):
    """Return the first quartile value (25/75) of roi"""
    return np.percentile(roi, 25)

def q3(roi):
    """Return the third quartile (75/25) value of roi"""
    return np.percentile(roi, 75)



# Private Functions (must begin with "_")
def _listStats():
    """
    Return list of the names of all non-private functions from acestats. 
    Private functions are excluded by their leading '_'.
    """
    from acerim import acestats
    all_func = np.array(inspect.getmembers(acestats, inspect.isfunction))
    stat_func = all_func[np.where([a[0][0]  != '_' for a in all_func])]
    return stat_func[:,0]

def _getFunctions(stats):
    """
    Return functions from this module according to stats. If stats is
    undefined, return all functions from this module, excluding private 
    functions.
    
    Returns
    -------
    List of lists containing 2 element pairs of function names and functions. 
        E.g. array( ['func name 1', <func1>], ['func name 2', <func2>])    
    """
    from acerim import acestats
    if isinstance(stats, str):
        stats = [stats]
    invalid_stats = [stat for stat in stats if stat not in _listStats()]
    if invalid_stats:
        raise ValueError('The following stats are not defined in acestats.py: '
                         + str(invalid_stats))
    all_func = inspect.getmembers(acestats, inspect.isfunction)
    stat_func = []
    for i, func in enumerate(all_func):
        if func[0] in stats:
            stat_func.append(all_func[i])
    return stat_func


""" 
Supplies instructions to setuptools to build acerim package distributions. 
Contains the metadata that appears on PyPI and automatically fetches version 
number from version.py. Also automatically builds docs using sphinx. 
"""
from setuptools import setup, find_packages
from codecs import open
from sphinx.setup_command import BuildDoc
from os import path
here = path.abspath(path.dirname(__file__))


def get_version():
    """ Fetch version from /acerim/version.py """
    version = {}
    with open(path.join(here, 'acerim', 'version.py')) as f:
        exec(f.read(), version)
    return version['__version__']

def get_readme():
    """ Fetch long winded description from /README.rst """
    with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
        return f.read()


# Set metadata
NAME = 'acerim'
VERSION = get_version()
DESCRIPTION = 'A package for analyzing impact crater ejecta.'
LONG_DESCRIPTION = get_readme()
AUTHOR = 'Christian Tai Udovicic'
AUTHOR_EMAIL = 'cj.taiudovicic@gmail.com'
MAINTAINER = AUTHOR
MAINTAINER_EMAIL = AUTHOR_EMAIL
URL = 'http://github.com/cjtu/acerim'
LICENSE = 'MIT'
CLASSIFIERS = ['Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3'
    ]
KEYWORDS = 'planetary science crater ejecta data analysis'
PLATFORMS = 'OS Independent'
PACKAGES = find_packages(exclude=['tests']) 
PACKAGE_DATA = {'': ['*.csv', '*.tif']}
INSTALL_REQUIRES = ['numpy', 'scipy', 'matplotlib', 'gdal', 'pandas']
#PYTHON_REQUIRES = '>=2.7, <=3.3'
EXTRAS_REQUIRE = None # {'test': ['pytest'], 'cite': ['duecredit'],}

# Setup Sphinx integration to automatically build documentation 
CMDCLASS = {'build_docs': BuildDoc}
COMMAND_OPTIONS = { # optional and override docs/conf.py settings
        'build_docs': {
            'project': ('setup.py', NAME),
            'version': ('setup.py', VERSION),
            'release': ('setup.py', VERSION),
            'source_dir': ('setup.py', './docs/source'),
            'build_dir': ('setup.py', './docs/build')}}

# Run setup() function with metadata above
setup(
    name=NAME,
    version=VERSION, 
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    maintainer=MAINTAINER,
    maintainer_email=MAINTAINER_EMAIL,     
    url=URL,
    license=LICENSE,
    classifiers=CLASSIFIERS,
    keywords=KEYWORDS,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    platforms=PLATFORMS,     
    packages=PACKAGES,
    package_data=PACKAGE_DATA,
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    #python_requires=PYTHON_REQUIRES,
    cmdclass=CMDCLASS, 
    command_options=COMMAND_OPTIONS,
)



# -*- coding: utf-8 -*-
"""
This file contains the core ACERIM classes, the AceDataset and CraterDataFrame.
For usage, see sample/tutorial.rst.
"""
from __future__ import division, print_function, absolute_import
import gdal
import numpy as np
import pandas as pd
from acerim import acefunctions as af

############################### ACEDATASET ####################################
class AceDataset(object):
    """
    Wraps the GDAL Dataset class to load and manipulate image data.

    Inputting a string will assume a data file path and open it using gdal.Open.
    If projection information is available (e.g. in a geotiff), AceDataset will 
    attempt to scrape it and populate its attributes automaticaly. Otherwise, 
    the geographic info can be manually overridden by setting the desired 
    instance variable attributes, or by supplying them as arguments during 
    initialization. See help(gdal.Dataset) for list of available attributes and 
    methods from GDAL.

    Note: AceDataset does not reproject input data and assumes the input is in
    simple cylindrical (Plate Caree) projection.
    
    Parameters
    ----------
    dataset : str or gdal.Dataset
        If str, assume filename compatible with gdal.Open(data). It is 
        recommended to pass an open simple cylindrical projected gdal.Dataset 
        object if other import or reprojection options are required. 
    nlat : int, float
        North latitude of dataset in (decimal) degrees.
    slat : int, float
        South latitude of dataset in (decimal) degrees.
    wlon : int, float
        West latitude of dataset in (decimal) degrees.
    elon : int, float
        East latitude of dataset in (decimal) degrees.
    radius : int, float
        Radius of planeary body in km.
    ppd : int, float
        Resolution of dataset in pixels/degree.
    *kwargs : ...
        Additional attributes to include in this instance of AceDataset, 
        accessible by the supplied keyword.
    
    >>> import os
    >>> f=os.path.dirname(os.path.abspath('__file__'))+'/tests/moon.tif'
    >>> ads = AceDataset(f, radius=1737)
    """
    def __init__(self, dataset, nlat=None, slat=None, wlon=None, elon=None, 
                 radius=None, ppd=None, **kwargs):
        """
        Initialize AceDataset object. See help(AceDataset) for correct 
        usage.
        """        
        if isinstance(dataset, str):
            self.gdalDataset = gdal.Open(dataset)
            if not self.gdalDataset:
                raise ImportError('Unable to open file. Check file path.')
        elif isinstance(dataset, gdal.Dataset):
            self.gdalDataset = dataset
        else:
            raise ImportError('Invalid input dataset')
        args = [nlat, slat, wlon, elon, radius, ppd]
        attrs = ['nlat','slat','wlon','elon','radius','ppd']
        # Attempt to read geospatial information with get_info
        dsinfo = self.get_info()
        for i,arg in enumerate(args):
            if arg is None: # Attempt to fill geospatial info automatically
                setattr(self, attrs[i], dsinfo[i])
            else: # If argument is supplied, override automatic get_info
                setattr(self, attrs[i], arg)
        # Add key-value attributes to object
        for key, value in kwargs.items():
            setattr(self, key, value)


    def __getattr__(self, name):
        """Redirects method and attribute calls to self.gdalDataset."""
        if name not in self.__dict__: # Redirect if not in AceDataset
            try:
                func = getattr(self.__dict__['gdalDataset'], name)
                if callable(func): # Call method
                    def gdalDataset_wrapper(*args, **kwargs):
                        return func(*args, **kwargs)
                    return gdalDataset_wrapper
                else: # Not callable so must be attribute
                    return func
            except AttributeError as e:
                raise AttributeError('Object has no attribute {}'.format(name))

    def __repr__(self):
        """Return string representation of AceDataset with all attribute info"""
        attrs = self.nlat, self.slat, self.wlon, self.elon, self.radius, self.ppd
        rep = 'AceDataset object with bounds '
        rep += '({}N, {}S), ({}E, {}E), radius {} km, and {} ppd resolution'.format(*attrs)
        return rep
    
    
    def calc_mpp(self, lat=0):
        """
        Return the ground resolution in meters/pixel at the given latitude, 
        calculated with the greatcircdist function in acefunctions.py.
        
        Parameters
        ----------
        lat: int, float
            Current latitude. Defaults to the equator (lat=0) if not specified. 
        
        Examples
        --------
        >>> import os
        >>> f = os.path.dirname(os.path.abspath('__file__'))+'/tests/moon.tif'
        >>> a = AceDataset(f, radius = 1737)
        >>> '{:.3f}'.format(a.calc_mpp())
        '7.579'
        >>> '{:.3f}'.format(a.calc_mpp(50))
        '4.872'
        """
        pixwidth = 1/self.ppd
        dist = af.greatcircdist(lat, 0, lat, pixwidth, self.radius)
        return dist
              

    def get_info(self):
        """
        Return list of georeferencing and projection information from the input
        data file if available. See help(gdal.Dataset) for compatible files
        for the GetGeoTransform method.
        
        Returns
        -------
        nlat : int, float
            North latitude from gdal.Dataset.GetProjection()
        slat : int, float
            South latitude calculated using resolution (degrees/pixel) from 
            gdal.Dataset.GetProjection() and the y-size of the image.
        wlon : int, float
            West latitude from gdal.Dataset.GetProjection()
        elon : int, float
            East latitude calculated using resolution (degrees/pixel) from 
            gdal.Dataset.GetProjection() and the x-size of the image.
        ppd : float
            Pixel resolution in pixels/degree from gdal.Dataset.GetProjection.
            
        Examples
        --------
        >>> import os
        >>> f = os.path.dirname(os.path.abspath('__file__'))+'/tests/moon.tif'
        >>> a = AceDataset(f)
        >>> a.get_info()
        (90.0, -90.0, -180.0, 180.0, 6378.137, 4.0)
        """
        xsize, ysize = self.RasterXSize, self.RasterYSize
        geotrans = self.GetGeoTransform()
        try:
            radius = 0.001*float(self.GetProjection().split(',')[3]) # Assume WKT format
        except IndexError:
            raise ImportError('Dataset radius not defined')
            radius = None
        wlon, dpp, nlat = (geotrans[0], geotrans[1], geotrans[3])
        elon = wlon + xsize*dpp
        slat = nlat - ysize*dpp
        ppd = 1/dpp
        return nlat, slat, wlon, elon, radius, ppd
       
        
    def is_global(self):
        """
        Check if self has 360 degrees of longitude.
        
        Examples
        --------
        >>> import os
        >>> f = os.path.dirname(os.path.abspath('__file__'))+'/tests/moon.tif'
        >>> a = AceDataset(f, 90, -90, 0, 360, 1737)
        >>> a.is_global()
        True
        """
        return abs(self.elon - self.wlon - 360) <= 0.0001
    
    
    def get_roi(self, lat, lon, rad, wsize=1, mask_crater=False, plot_roi=False, 
               get_extent=False):
        """
        Return square 2D numpy array containing region of interest (ROI) 
        centered on (lat,lon). The window size is given by 2*wsize*rad and 
        extends wsize crater radii from the crater center. 

        The crater at the center given by the ellipse at lat, lon, rad is
        excluded from the ROI with the mask_crater flag. This replaces pixels 
        in the crater rim with NaN.
        
        If the requested ROI crosses the lon extent of a global dataset, use 
        wrap_lon() to concatenate the parts of the roi on either side of the 
        boundary. Otherwise, raise error if lat or lon out of bounds.
    
        Parameters
        ----------
        lat : int, float
            Crater latitude, centre latitude of ROI.
        lon : int, float
            Crater longitude, centre longitude of ROI.
        rad : int, float
            Crater radius.
        wsize : int, float
            Window size in crater radii. Side length of ROI around crater is 
            2*wsize*rad (E.g., ROI with wsize=1 gives smallest square around
            crater, making the side lengths 1 diam).
        mask_crater : bool
            Masks the crater floor from the resulting ROI by replacing the 
            crater with an ellipse of NaN.
        plot_roi : bool
            Plots the returned ROI.
        get_extent : bool
            Reuturns the ROI window extent as (minlon, maxlon, minlat, maxlat)
        
        Returns
        --------
        roi: 2Darray
            Numpy 2D array of data centered on the specified crater and 
            extending wsize*rad distance from the crater centre.
            
        roi, extent : (2Darray, tuple)
            If get_extent flag is True, return both the 2D roi and the extent
            tuple.
        """
        def wrap_lon(ads, minlon, maxlon, topind, height):
            """
            Return ROI that extends past the right edge of a global dataset by 
            wrapping and concatenating the right and left sides of the ROI.
            """
            if minlon < ads.wlon: 
                leftind = af.get_ind(minlon, lonarr - 360)
                leftwidth = af.deg2pix(ads.wlon - minlon, ads.ppd)
                rightind = af.get_ind(ads.wlon, lonarr)
                rightwidth = af.deg2pix(maxlon - ads.wlon, ads.ppd)
            elif maxlon > ads.elon:
                leftind = af.get_ind(minlon, lonarr)
                leftwidth = af.deg2pix(ads.elon - minlon, ads.ppd)
                rightind = af.get_ind(ads.elon, lonarr + 360)
                rightwidth = af.deg2pix(maxlon - ads.elon, ads.ppd)                
            left_roi = ads.ReadAsArray(leftind, topind, leftwidth, height)
            right_roi = ads.ReadAsArray(rightind, topind, rightwidth, height)
            return np.concatenate((left_roi, right_roi), axis=1)
                      
        # If crater lon out of bounds, switch lon convention [(0,360) <-> (-180,180)]
        if lon > self.elon: 
            lon -= 360
        if lon < self.wlon:
            lon += 360
        # Get window extent in degrees
        dwsize = af.m2deg(wsize*rad, self.calc_mpp(lat), self.ppd)
        minlat = lat-dwsize
        maxlat = lat+dwsize
        minlon = lon-dwsize
        maxlon = lon+dwsize
        extent = (minlon, maxlon, minlat, maxlat)
        # Throw error if window bounds are not in lat bounds.
        if minlat < self.slat or maxlat > self.nlat:
            raise ImportError('Latitude ({},{}) out of dataset bounds ({},{}) '.format(
                                       minlat, maxlat, self.slat, self.nlat))
        # Make dataset pixel arrays
        latarr = np.linspace(self.nlat, self.slat, self.RasterYSize)
        lonarr = np.linspace(self.wlon, self.elon, self.RasterXSize)
        # Get top index and height of ROI
        topind = af.get_ind(maxlat, latarr) 
        height = af.deg2pix(2*dwsize, self.ppd)
        # Check if ROI crosses lon extent of global ROI and needs to be wrapped
        if self.is_global() and (minlon < self.wlon or maxlon > self.elon):
            roi = wrap_lon(self, minlon, maxlon, topind, height)  
        else: 
            # Get left index and width of ROI and read data from dataset
            leftind = af.get_ind(minlon,lonarr) 
            width = af.deg2pix(2*dwsize, self.ppd)
            roi = self.ReadAsArray(leftind, topind, width, height) # gdal Dataset method
        if roi is None:
            raise ImportError('GDAL could not read dataset into array')
            return
        if mask_crater:
            cmask = af.crater_floor_mask(self, roi, lat, lon, rad)
            roi = af.mask_where(roi, cmask)
        if plot_roi:
            self.plot_roi(roi, extent=extent)    
        if get_extent:
            return roi, extent
        else:
            return roi 

    
    def plot_roi(self, roi, *args, **kwargs):
        """
        Implements plot_roi function in functions.py. 
        """
        af.plot_roi(self, roi, *args, **kwargs)


############################## CRATERDATAFRAME ################################
class CraterDataFrame(pd.DataFrame):
    """
    Extends DataFrame from the Pandas module. DataFrames are "two-dimensional 
    size-mutable, potentially heterogeneous tabular data structures". They are
    a convenient way to contain and manipulate tabular data in Python.

    The CraterDataFrame differs from the stock DataFrame in that it detects the
    columns of latitude, longitude, and radius (or diameter),
    storing a reference to each column in latcol, loncol and radcol,
    respectively. This ensures that crater location data can be extracted from 
    the correct data columns. It also allows the option to initialize using from 
    a csv filename or by passing a pandas.DataFrame object with Lat, Lon, and 
    size columns.

    Sicne CraterDataFrame inherits from pandas.DataFrame, all DataFrame methods
    are available for use in the CraterDataFrame. See help(pandas.DataFrame)
    for methods and DataFrame usage. 
    
    Parameters
    ----------
    data : str or pandas.DataFrame
        Str will assume a filename ending in '.csv' with data to be read. It is 
        recommended to pass a pandas.DataFrame object if complicated import 
        options are required.
    **kwargs : 
        All optional arguments from pandas.DataFrame can be appended as keyword 
        arguments.    
        
    Examples
    --------
    >>> cdict = {'Lat' : [10, -20., 80.0], 
                 'Lon' : [14, -40.1, 317.2],
                 'Diam' : [2, 12., 23.7]}
    >>> cdf = CraterDataFrame(cdict)
    >>> cdf['Diam'][0]
    2.0
    >>> index = ['Crater A', 'Crater B', 'Crater C']
    >>> cdf2 = CraterDataFrame(cdict, index=index)
    >>> cdf2.loc['Crater A']['Lat']
    10.0
    >>> cdf2.latcol
    'Lat'
    """
    def __init__(self, data=None, index_col=None, latcol=None, 
    			 loncol=None, radcol=None, **kwargs):
        """
        Initialize a CraterDataFrame object with data (a str filename, a 
        pandas.DataFrame object, or any of the acceptable inputs to 
        pandas.DataFrame). See help(CraterDataFrame) for correct usage.
        """
        if isinstance(data, str):
            data = pd.read_csv(data, index_col=index_col) 
        super(CraterDataFrame, self).__init__(data, **kwargs)
        
        # If no lat-, lon-, or rad- col provided, try to find them in columns
        colnames = ['latitude', 'longitude', 'radius']
        colabbrevs = ['lat', 'lon', 'rad']
        attrs = [None, None, None]

        for i in range(len(colnames)):
            if not attrs[i]: # If not defined in constructor
                findcol = [(colnames[i] == col.strip().lower()) or 
                            (colabbrevs[i] == col.strip().lower()) for col in self.columns]
                if any(findcol):
                    attrs[i] = self.columns[np.where(findcol)[0][0]]
                elif colnames[i] == 'radius':
                    finddiam = [('diameter' == col.strip().lower()) or 
                                ('diam' == col.strip().lower()) for col in self.columns]    
                    if any(finddiam):
                        diamcol = self.columns[np.where(finddiam)[0][0]]
                        self['radius'] = (0.5)*self[diamcol]
                        attrs[i] = 'radius'                    
                if not attrs[i]:
                    raise ImportError('Unable to infer {} column from header. '.format(colnames[i])+
                                      'Specify {}col in constructor.'.format(colabbrevs[i]))

        self.latcol = attrs[0]
        self.loncol = attrs[1]
        self.radcol = attrs[2]


if __name__ == "__main__":
    import doctest
    doctest.testmod()
